''' ------------------   Data and functions importation   ------------------ '''

# Data importation
from DataStructure import aircraft, phase, service, vehicle, environment, solution
AC = aircraft()
P = phase()
S = service()
V = vehicle()
env = environment()
sol = solution()

# Functions importation
from SelectVehicle import SelectVehicle
from HandleDependencies import dependence
from PerformanceMeasures import ServiceTime, Lateness, Buffer

# Librairies importation
import datetime as dt
import time

''' -------------------------------   Main   ------------------------------- '''

## Sort AC following a chosen strategy

#Strategy 1: Aircraft ID are sorted with the FIFO method (by order of arrivals)
AC_1 = AC.ID.copy()

#Srategy 2: ...
#AC_2 = ...

#Srategy 3: ...
#AC_3 = ...

start = time.time()
## Create the solution
#Consider AC one by one following the order given by the previously chosen sorting strategy
for i in AC_1:
    #print("--------- AC =", i, "---------")
    
    #Consider each phase (group of services) successively
    for p in range (0, P.nb):
        #print(" --- Phase ", p, "---")
        
        #Initialization of t and tmax to the earliest start time of phase p at AC i
        tmax = P.a[p][i]
        t = P.a[p][i]
        
        #Consider phase p starting services one by one
        for s in S.S[p]:
            #Initialization of q_0 to the total quantity that should be served for service s at AC i
            q_0 = S.q[s][i]
            #Initialization of tsuite to the earliest start time of phase p at AC i
            tsuite = t
            
            #While AC i is not completely served for service s, we do not start another starting service
            while S.q[s][i] > 0:
                #Vehicle selection
                #If service s can be operated by several vehicles at once
                if S.simultaneity[s] == True:
                    #The selected vehicle can start the service at t (no waiting time)
                    v_best, load, t_end, t_new = SelectVehicle(i, s, 0, t, S.q[s][i], AC, env, S, V, sol)
                    
                #If service s cannot be operated by several vehicles at once
                else:
                    #The selected vehicle can start only when the previously selected vehicle 
                    #for service s has finished its task (cannot start before tsuite)
                    v_best, load, t_end, t_new = SelectVehicle(i, s, 0, tsuite, S.q[s][i], AC, env, S, V, sol)
                
                #Update the quantity that remains to be served and time parameters
                S.q[s][i] -= load
                tmax = max(t_end, tmax)
                tsuite = t_end
                sol.EndTime[i][s] = t_end
                
                #If s is the head of a transfer chain OR if s has a precedence constraint with another service
                if S.QGlobal[s] == False or S.q[s][i] == 0:
                    t_end, qremain = dependence(i, s, v_best, q_0, t_end, load, S, V, AC, env, sol)
                
                #Load has gone through the whole transfer chain, the global quantity q_0 can thus be reduced accordingly
                q_0 = qremain
                
                #Update time parameters
                tmax = max(t_end, tmax)
               
            #Once the service completed, authorize vehicles to start another service
            for v in range(0, V.nb):
                
                #Make it available again
                if V.used[v] == True:
                    V.used[v] = False
                    
                    #If the considered vehicle must follow a services sequence (dollies only)
                    if len(V.sequence[v]) > 0:
                        #Dollies can follow 2 different services sequences
                        #If the dollies are involved in the 1st services sequence
                        if V.currentService[v] in V.sequence[v][0]:
                            x = V.sequence[v][0].index(V.currentService[v])
                            #If we are not at the end of the services sequence, 
                            #go to the next service of the sequence
                            if x < len(V.sequence[v][0]) - 1:
                                y = x + 1
                                V.currentService[v] = V.sequence[v][0][y]
                            #Else, dollies can perform any services of its service list
                            else:
                                V.currentService[v] = "any"
                        #If the dollies are involved in the 2nd services sequence
                        else:
                            x = V.sequence[v][1].index(V.currentService[v])
                            #If we are not at the end of the services sequence, 
                            #go to the next service of the sequence
                            if x < len(V.sequence[v][1]) - 1:
                                y = x + 1
                                V.currentService[v] = V.sequence[v][1][y]
                            #Else, dollies can perform any services of its service list
                            else:
                                V.currentService[v] = "any"
                    #If the considered vehicle has no services sequence to follow
                    else:
                        #It can perform any services of its service list
                        V.currentService[v] = "any"
                    
                    #If the vehicle v is not involved in a service sequence (anymore),
                    #it can serve again any compatible AC
                    if V.currentService[v] == "any":
                        V.compatibleAC[v] = V.compatibleAC_init[v].copy()
                                        
        #Update earliest start time of following phases taking into account precedence constraints
        for q in range(p+1, P.nb):
            #If phase p should be completed before phase q can start
            if P.precedenceMatrix[q,p] == 1:
                #We can start to bring outgoing ULDs on AC parking short before the end of unloading
                if q == 4:
                    P.a[q][i] = max(P.a[q][i], (tmax - dt.timedelta(minutes=30)))
                #Update the earliest start time of phase q 
                #(only if tmax is higher than the current earliest start time of q)
                else:
                    P.a[q][i] = max(P.a[q][i], tmax)

end = time.time()

## Compute objective function measures
#Initialization
TotalST = dt.timedelta(seconds=0)
WorstST = dt.timedelta(seconds=0)
MinST = dt.timedelta(seconds=86400)
last_t = [dt.timedelta(seconds=0) for a in AC.ID]

#Service time
for i in AC.ID:
    #Compute individual service time for each aircraft
    ST, last_t[i] = ServiceTime(i, AC, sol)

    #Compute the worst case individual service time
    if ST > WorstST:
        WorstST = ST

    #Compute the minimum individual service time
    if ST < MinST:
        MinST = ST
    
    #Compute total service time
    TotalST += ST


print("Min individual service time is ", MinST)
print("Max individual service time is ", WorstST)

#Compute the mean service time
MeanST = TotalST / len(AC.ID)
print("Mean service time is ", MeanST)

#Compute AC delay
print("-------------------------------------------------------")
Lateness(AC, sol)

#Compute AC buffer before TWb
print("-------------------------------------------------------")
Buffer(AC, sol)

#Compute computing time
print("-------------------------------------------------------")
CTime = end - start
print("Computing time : ", CTime)




